@@ -1,5 +1,5 @@
 /* pinentry.c - The PIN entry support library
-   Copyright (C) 2002, 2003, 2007 g10 Code GmbH
+   Copyright (C) 2002, 2003, 2007, 2008, 2010 g10 Code GmbH
    
    This file is part of PINENTRY.
    
@@ -14,24 +14,30 @@
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA  */
+   along with this program; if not, see <http://www.gnu.org/licenses/>.  
+ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
-#include <errno.h>
+#ifndef HAVE_W32CE_SYSTEM
+# include <errno.h>
+#endif
 #include <stdlib.h>
 #include <string.h>
 #include <getopt.h>
 #include <unistd.h>
-#include <locale.h>
+#ifndef HAVE_W32CE_SYSTEM
+# include <locale.h>
+#endif
 #ifdef HAVE_LANGINFO_H
 #include <langinfo.h>
 #endif
 #include <limits.h>
+#ifdef HAVE_W32CE_SYSTEM
+# include <windows.h>
+#endif
 
 #if defined FALLBACK_CURSES || defined PINENTRY_CURSES || defined PINENTRY_GTK
 #include <iconv.h>
@@ -42,6 +48,9 @@
 #include "secmem-util.h"
 #include "pinentry.h"
 
+#ifdef HAVE_W32CE_SYSTEM
+#define getpid() GetCurrentProcessId ()
+#endif
 
 /* Keep the name of our program here. */
 static char this_pgmname[50]; 
@@ -49,10 +58,12 @@ static char this_pgmname[50];
 
 struct pinentry pinentry =
   {
+    NULL,	/* Title.  */
     NULL,	/* Description.  */
     NULL,	/* Error.  */
     NULL,	/* Prompt.  */
     NULL,	/* Ok button.  */
+    NULL,	/* Not-Ok button.  */
     NULL,	/* Cancel button.  */
     NULL,	/* PIN.  */
     2048,	/* PIN length.  */
@@ -62,11 +73,15 @@ struct pinentry pinentry =
     0,		/* TTY LC_CTYPE.  */
     0,		/* TTY LC_MESSAGES.  */
     0,		/* Debug mode.  */
+#ifdef ENABLE_ENHANCED
     0,		/* Enhanced mode.  */
+#endif
     1,		/* Global grab.  */
     0,		/* Parent Window ID.  */
     NULL,       /* Touch file.  */
     0,		/* Result.  */
+    0,		/* Canceled.  */
+    0,		/* Close button flag.  */
     0,          /* Locale error flag. */
     0,          /* One-button flag.  */
     NULL,       /* Quality-Bar flag and description.  */
@@ -76,11 +91,13 @@ struct pinentry pinentry =
     PINENTRY_COLOR_DEFAULT,
     PINENTRY_COLOR_DEFAULT,
     0,
-    NULL        /* Assuan context.  */
+    NULL,        /* default_ok  */
+    NULL,        /* default_cancel  */
+    NULL,        /* default_prompt  */
+    NULL         /* Assuan context.  */
   };
 
 
-
 #if defined FALLBACK_CURSES || defined PINENTRY_CURSES || defined PINENTRY_GTK
 char *
 pinentry_utf8_to_local (char *lc_ctype, char *text)
@@ -370,13 +387,15 @@ pinentry_init (const char *pgmname)
 int
 pinentry_have_display (int argc, char **argv)
 {
+#ifndef HAVE_W32CE_SYSTEM
   const char *s;
 
   s = getenv ("DISPLAY");
   if (s && *s)
     return 1;
+#endif
   for (; argc; argc--, argv++)
-    if (!strcmp (*argv, "--display"))
+    if (!strcmp (*argv, "--display") || !strncmp (*argv, "--display=", 10))
       return 1;
   return 0;
 }
@@ -394,7 +413,9 @@ usage (void)
 "      --ttytype NAME    Set the tty terminal type\n"
 "      --lc-ctype        Set the tty LC_CTYPE value\n"
 "      --lc-messages     Set the tty LC_MESSAGES value\n"
+#ifdef ENABLE_ENHANCED
 "  -e, --enhanced        Ask for timeout and insurance, too\n"
+#endif
 "  -g, --no-global-grab  Grab keyboard only while window is focused\n"
 "      --parent-wid	 Parent window ID (for positioning)\n"
 "  -d, --debug           Turn on debugging output\n"
@@ -469,7 +490,9 @@ pinentry_parse_opts (int argc, char *argv[])
      { "ttytype", required_argument,     0, 'N' },
      { "lc-ctype", required_argument,    0, 'C' },
      { "lc-messages", required_argument, 0, 'M' },
+#ifdef ENABLE_ENHANCED
      { "enhanced", no_argument,          0, 'e' },
+#endif
      { "no-global-grab", no_argument,    0, 'g' },
      { "parent-wid", required_argument,  0, 'W' },
      { "colors", required_argument,	 0, 'c' },
@@ -487,9 +510,11 @@ pinentry_parse_opts (int argc, char *argv[])
         case 'd':
           pinentry.debug = 1;
           break;
+#ifdef ENABLE_ENHANCED
         case 'e':
           pinentry.enhanced = 1;
           break;
+#endif
         case 'g':
           pinentry.grab = 0;
           break;
@@ -503,7 +528,9 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.display = strdup (optarg);
 	  if (!pinentry.display)
 	    {
+#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break; 
@@ -511,7 +538,9 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.ttyname = strdup (optarg);
 	  if (!pinentry.ttyname)
 	    {
+#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -519,7 +548,9 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.ttytype = strdup (optarg);
 	  if (!pinentry.ttytype)
 	    {
+#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -527,7 +558,9 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.lc_ctype = strdup (optarg);
 	  if (!pinentry.lc_ctype)
 	    {
+#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -535,7 +568,9 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.lc_messages = strdup (optarg);
 	  if (!pinentry.lc_messages)
 	    {
+#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -577,10 +612,12 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
     pinentry.grab = 1;
   else if (!strcmp (key, "debug-wait"))
     {
+#ifndef HAVE_W32CE_SYSTEM
       fprintf (stderr, "%s: waiting for debugger - my pid is %u ...\n",
 	       this_pgmname, (unsigned int) getpid());
       sleep (*value?atoi (value):5);
       fprintf (stderr, "%s: ... okay\n", this_pgmname);
+#endif
     }
   else if (!strcmp (key, "display"))
     {
@@ -635,6 +672,24 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
       if (!pinentry.touch_file)
 	return ASSUAN_Out_Of_Core;
     }
+  else if (!strcmp (key, "default-ok"))
+    {
+      pinentry.default_ok = strdup (value);
+      if (!pinentry.default_ok)
+	return ASSUAN_Out_Of_Core;
+    }
+  else if (!strcmp (key, "default-cancel"))
+    {
+      pinentry.default_cancel = strdup (value);
+      if (!pinentry.default_cancel)
+	return ASSUAN_Out_Of_Core;
+    }
+  else if (!strcmp (key, "default-prompt"))
+    {
+      pinentry.default_prompt = strdup (value);
+      if (!pinentry.default_prompt)
+	return ASSUAN_Out_Of_Core;
+    }
   else
     return ASSUAN_Invalid_Option;
   return 0;
@@ -644,7 +699,7 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
 /* Note, that it is sufficient to allocate the target string D as
    long as the source string S, i.e.: strlen(s)+1; */
 static void
-strcpy_escaped (char *d, const unsigned char *s)
+strcpy_escaped (char *d, const char *s)
 {
   while (*s)
     {
@@ -729,6 +784,23 @@ cmd_setok (ASSUAN_CONTEXT ctx, char *line)
 }
 
 
+static int
+cmd_setnotok (ASSUAN_CONTEXT ctx, char *line)
+{
+  char *newo;
+  newo = malloc (strlen (line) + 1);
+
+  if (!newo)
+    return ASSUAN_Out_Of_Core;
+
+  strcpy_escaped (newo, line);
+  if (pinentry.notok)
+    free (pinentry.notok);
+  pinentry.notok = newo;
+  return 0;
+}
+
+
 static int
 cmd_setcancel (ASSUAN_CONTEXT ctx, char *line)
 {
@@ -746,6 +818,21 @@ cmd_setcancel (ASSUAN_CONTEXT ctx, char *line)
 }
 
 
+static int
+cmd_settitle (ASSUAN_CONTEXT ctx, char *line)
+{
+  char *newt;
+  newt = malloc (strlen (line) + 1);
+  
+  if (!newt)
+    return ASSUAN_Out_Of_Core;
+  
+  strcpy_escaped (newt, line);
+  if (pinentry.title)
+    free (pinentry.title);
+  pinentry.title = newt;
+  return 0;
+}
 
 static int
 cmd_setqualitybar (ASSUAN_CONTEXT ctx, char *line)
@@ -800,10 +887,11 @@ cmd_getpin (ASSUAN_CONTEXT ctx, char *line)
     return ASSUAN_Out_Of_Core;
   if (!pinentry.prompt)
     {
-      pinentry.prompt = "PIN:";
+      pinentry.prompt = pinentry.default_prompt?pinentry.default_prompt:"PIN:";
       set_prompt = 1;
     }
   pinentry.locale_err = 0;
+  pinentry.close_button = 0;
   pinentry.one_button = 0;
   pinentry.ctx_assuan = ctx;
   result = (*pinentry_cmd_handler) (&pinentry);
@@ -818,6 +906,9 @@ cmd_getpin (ASSUAN_CONTEXT ctx, char *line)
 
   pinentry.quality_bar = 0;  /* Reset it after the command.  */
 
+  if (pinentry.close_button)
+    assuan_write_status (ctx, "BUTTON_INFO", "close");
+
   if (result < 0)
     {
       if (pinentry.pin)
@@ -845,7 +936,7 @@ cmd_getpin (ASSUAN_CONTEXT ctx, char *line)
 }
 
 
-/* Note that the option --one-button is hack to allow the use of old
+/* Note that the option --one-button is a hack to allow the use of old
    pinentries while the caller is ignoring the result.  Given that
    options have never been used or flagged as an error the new option
    is an easy way to enable the messsage mode while not requiring to
@@ -859,7 +950,9 @@ cmd_confirm (ASSUAN_CONTEXT ctx, char *line)
 
   pinentry.one_button = !!strstr (line, "--one-button");
   pinentry.quality_bar = 0;
+  pinentry.close_button = 0;
   pinentry.locale_err = 0;
+  pinentry.canceled = 0;
   result = (*pinentry_cmd_handler) (&pinentry);
   if (pinentry.error)
     {
@@ -867,11 +960,16 @@ cmd_confirm (ASSUAN_CONTEXT ctx, char *line)
       pinentry.error = NULL;
     }
 
+  if (pinentry.close_button)
+    assuan_write_status (ctx, "BUTTON_INFO", "close");
+
   return result ? 0
                 : (pinentry.locale_err? ASSUAN_Locale_Problem
                                       : (pinentry.one_button 
                                          ? 0
-                                         : ASSUAN_Not_Confirmed));
+                                         : (pinentry.canceled
+                                            ? ASSUAN_Canceled
+                                            : ASSUAN_Not_Confirmed)));
 }
 
 
@@ -882,6 +980,7 @@ cmd_message (ASSUAN_CONTEXT ctx, char *line)
 
   pinentry.one_button = 1;
   pinentry.quality_bar = 0;
+  pinentry.close_button = 0;
   pinentry.locale_err = 0;
   result = (*pinentry_cmd_handler) (&pinentry);
   if (pinentry.error)
@@ -890,11 +989,44 @@ cmd_message (ASSUAN_CONTEXT ctx, char *line)
       pinentry.error = NULL;
     }
 
+  if (pinentry.close_button)
+    assuan_write_status (ctx, "BUTTON_INFO", "close");
+
   return result ? 0 
                 : (pinentry.locale_err? ASSUAN_Locale_Problem
                                       : 0);
 }
 
+/* GETINFO <what>
+
+   Multipurpose function to return a variety of information.
+   Supported values for WHAT are:
+
+     version     - Return the version of the program.
+     pid         - Return the process id of the server.
+ */
+static int
+cmd_getinfo (assuan_context_t ctx, char *line)
+{
+  int rc;
+
+  if (!strcmp (line, "version"))
+    {
+      const char *s = VERSION;
+      rc = assuan_send_data (ctx, s, strlen (s));
+    }
+  else if (!strcmp (line, "pid"))
+    {
+      char numbuf[50];
+
+      snprintf (numbuf, sizeof numbuf, "%lu", (unsigned long)getpid ());
+      rc = assuan_send_data (ctx, numbuf, strlen (numbuf));
+    }
+  else
+    rc = ASSUAN_Parameter_Error;
+  return rc;
+}
+
 
 /* Tell the assuan library about our commands.  */
 static int
@@ -911,12 +1043,15 @@ register_commands (ASSUAN_CONTEXT ctx)
       { "SETPROMPT",  0,  cmd_setprompt },
       { "SETERROR",   0,  cmd_seterror },
       { "SETOK",      0,  cmd_setok },
+      { "SETNOTOK",   0,  cmd_setnotok },
       { "SETCANCEL",  0,  cmd_setcancel },
       { "GETPIN",     0,  cmd_getpin },
       { "CONFIRM",    0,  cmd_confirm },
       { "MESSAGE",    0,  cmd_message },
       { "SETQUALITYBAR", 0,  cmd_setqualitybar },
       { "SETQUALITYBAR_TT", 0,  cmd_setqualitybar_tt },
+      { "GETINFO",    0,  cmd_getinfo },
+      { "SETTITLE",   0,  cmd_settitle },
       { NULL }
     };
   int i, j, rc;
@@ -934,11 +1069,8 @@ register_commands (ASSUAN_CONTEXT ctx)
 }
 
 
-/* Start the pinentry event loop.  The program will start to process
-   Assuan commands until it is finished or an error occurs.  If an
-   error occurs, -1 is returned.  Otherwise, 0 is returned.  */
 int
-pinentry_loop (void)
+pinentry_loop2 (int infd, int outfd)
 {
   int rc;
   int filedes[2];
@@ -953,8 +1085,8 @@ pinentry_loop (void)
   /* For now we use a simple pipe based server so that we can work
      from scripts.  We will later add options to run as a daemon and
      wait for requests on a Unix domain socket.  */
-  filedes[0] = 0;
-  filedes[1] = 1;
+  filedes[0] = infd;
+  filedes[1] = outfd;
   rc = assuan_init_pipe_server (&ctx, filedes);
   if (rc)
     {
@@ -999,3 +1131,13 @@ pinentry_loop (void)
   assuan_deinit_server (ctx);
   return 0;
 }
+
+
+/* Start the pinentry event loop.  The program will start to process
+   Assuan commands until it is finished or an error occurs.  If an
+   error occurs, -1 is returned.  Otherwise, 0 is returned.  */
+int
+pinentry_loop (void)
+{
+  return pinentry_loop2 (STDIN_FILENO, STDOUT_FILENO);
+}
